#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# TODO: improve this computation.
# Probably bit reversal is not the best way to do it.

def crc(data):
    poly = 0x8408
    crc = 0xffff
    for byte in data:
        # reverse byte
        byte = (byte & 0x55) << 1 | (byte & 0xAA) >> 1
        byte = (byte & 0x33) << 2 | (byte & 0xCC) >> 2
        byte = (byte & 0x0F) << 4 | (byte & 0xF0) >> 4
        crc ^= byte
        for _ in range(8):
            crc = (crc >> 1) ^ poly if crc & 1 else crc >> 1
    crc = (crc & 0x5555) << 1 | (crc & 0xAAAA) >> 1
    crc = (crc & 0x3333) << 2 | (crc & 0xCCCC) >> 2
    crc = (crc & 0x0F0F) << 4 | (crc & 0xF0F0) >> 4
    crc = (crc & 0x00FF) << 8 | (crc & 0xFF00) >> 8
    return crc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from .imagereceiver import ImageReceiver


block_size = 46

swiatowid_image_block = Struct(
    'sequence' / Int16ul,
    'data' / Bytes(block_size)
    )


class ImageReceiverSwiatowid(ImageReceiver):
    def filename(self, fid):
        return f'{fid}.jpg'

    def chunk_size(self):
        return block_size

    def parse_chunk(self, chunk):
        if len(chunk) != swiatowid_image_block.sizeof():
            return None
        try:
            frame = swiatowid_image_block.parse(chunk)
        except ConstructError:
            return None
        return frame


swiatowid = ImageReceiverSwiatowid



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, digital
import numpy as np
import pmt

from ... import viterbi_decoder
from ...grtypes import byte_t
from ...hdlc_deframer import hdlc_deframer
from ...hier.sync_to_pdu import sync_to_pdu
from ...utils.options_block import options_block
from ...check_tt64_crc import crc_table


# Two HDLC flags encoded and interleaved (32 bits)
_syncword = '10001010111001101000101011100110'


class deinterleave(gr.basic_block):
    """
    Helper block to perform 4x4 matrix deinterleaving

    The input and output are PDUs with unpacked bits
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='deinterleave',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = np.array(pmt.u8vector_elements(msg), dtype='uint8')
        if packet.size % 16 != 0:
            print('[ERROR] Packet size is not a multiple of 16 bits')
            return
        packet = np.einsum('ijk->ikj', packet.reshape((-1, 4, 4))).ravel()
        # Invert every other bit, since the g0 branch of the convolutional
        # encoder is inverted.
        packet[::2] ^= 1
        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(packet), packet)))


def crc16_usb(data):
    crc = 0xffff
    for byte in data:
        tbl_idx = (crc ^ byte) & 0xff
        crc = (crc_table[tbl_idx] ^ (crc >> 8)) & 0xffff
    return (crc ^ 0xffff) & 0xffff


def crc_check(frame):
    if len(frame) <= 2:
        return False
    crc = crc16_usb(frame[:-2])
    return frame[-2] == (crc & 0xff) and frame[-1] == ((crc >> 8) & 0xff)


class ax5043_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe AX5043 FEC packets

    The input is a float stream of soft symbols. The output are PDUs
    with the FEC frames. The AX5043 uses a k = 5, r = 1/2 convolutional
    code, 4x4 matrix interleaving, and HDLC framing. This deframer uses
    a Viterbi decoder that works on hard-decision bits.

    Args:
        options: Options from argparse
    """
    def __init__(self, options=None):
        gr.hier_block2.__init__(
            self,
            'ax5043_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        self.slicer = digital.binary_slicer_fb()
        # 4000 bits will leave enough room for the 200 byte packets
        self.deframer = sync_to_pdu(
            packlen=4000, sync=_syncword, threshold=4)
        self.deinterleave = deinterleave()
        self.viterbi = viterbi_decoder(5, [25, 23])
        self.pdu2tag = blocks.pdu_to_tagged_stream(byte_t, 'packet_len')
        self.hdlc = hdlc_deframer(True, 10000, crc_check_func=crc_check)

        self.connect(self, self.slicer, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.deinterleave, 'in'))
        self.msg_connect((self.deinterleave, 'out'), (self.viterbi, 'in'))
        self.msg_connect((self.viterbi, 'out'), (self.pdu2tag, 'pdus'))
        self.connect(self.pdu2tag, self.hdlc)
        self.msg_connect((self.hdlc, 'out'), (self, 'out'))

    @classmethod
    def add_options(cls, parser):
        """
        Adds AX5043 deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--verbose_crc', action='store_true', help='Verbose CRC')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import hashlib

from gnuradio import gr
import numpy
import pmt
import requests


class funcube_submit(gr.basic_block):
    """docstring for block funcube_submit"""
    def __init__(self, url, site_id, auth_code):
        gr.basic_block.__init__(
            self,
            name='funcube_submit',
            in_sig=[],
            out_sig=[])

        self.base_url = url
        self.site_id = site_id
        self.auth_code = auth_code

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def digest(self, hex_string):
        md5 = hashlib.md5()
        md5.update(hex_string)
        md5.update(b':')
        md5.update(bytes(self.auth_code, encoding='ascii'))
        return md5.digest().hex()

    def handle_msg(self, msg_pmt):
        if not self.base_url or not self.site_id or not self.auth_code:
            return

        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return

        frame = bytes(bytes(pmt.u8vector_elements(msg)).hex().upper(),
                      encoding='ascii')
        # Add a space every two hex chars
        frame = b' '.join([frame[j:j+2] for j in range(0, len(frame), 2)])

        url = (self.base_url + '/api/data/hex/'
               + self.site_id + '/?digest=' + self.digest(frame))
        r = requests.post(url, data=frame,
                          headers={'Content-Type': 'application/text'})
        if r.status_code != 200:
            print('FUNcube server error while submitting telemetry')
            print('Reply:', r.text)
            print('HTTP code', r.status_code)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import collections

from gnuradio import gr
import numpy
import pmt

from . import hdlc


class hdlc_framer(gr.basic_block):
    """docstring for block hdlc_framer"""
    def __init__(self, preamble_bytes, postamble_bytes):
        gr.basic_block.__init__(
            self,
            name='hdlc_framer',
            in_sig=None,
            out_sig=None)

        self.preamble_bytes = preamble_bytes
        self.postamble_bytes = postamble_bytes
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return

        data = list(pmt.u8vector_elements(msg))
        crc = hdlc.crc_ccitt(data)
        data.append(crc & 0xff)
        data.append((crc >> 8) & 0xff)

        buff = list(hdlc.flag * self.preamble_bytes)
        ones = 0  # number of consecutive ones
        for byte in data:
            for _ in range(8):
                # Transmit byte LSB first
                x = byte & 1
                buff.append(x)
                if x:
                    ones += 1
                else:
                    ones = 0
                if ones == 5:
                    # Bit-stuff
                    buff.append(0)
                    ones = 0
                byte >>= 1
        buff.extend(hdlc.flag * self.postamble_bytes)

        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(buff), buff)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from .ax25 import Header


amicalsat = Struct(
        'header' / Header,
        'info' / GreedyString('ascii')
        )



# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: CCSDS/NASA-GSFC Viterbi decoder
# Author: Daniel Estevez
# Description: Viterbi27 decoder with convention (POLYB, ~POLYA)
#              (output is unpacked)
# GNU Radio version: 3.8.0.0

from gnuradio import fec
from gnuradio import gr
from gnuradio.filter import firdes


class ccsds_viterbi(gr.hier_block2):
    def __init__(self, code='CCSDS'):
        gr.hier_block2.__init__(
            self,
            'CCSDS/NASA-GSFC Viterbi decoder',
            gr.io_signature(1, 1, gr.sizeof_float*1),
            gr.io_signature(1, 1, gr.sizeof_char*1),
        )

        ##################################################
        # Variables
        ##################################################
        polys = {'CCSDS': [79, -109],
                 'NASA-DSN': [-109, 79],
                 'CCSDS uninverted': [79, 109],
                 'NASA-DSN uninverted': [109, 79],
                 }
        self.dec_cc = dec_cc = fec.cc_decoder.make(
            80, 7, 2, polys[code], 0, -1, fec.CC_STREAMING, False)

        ##################################################
        # Blocks
        ##################################################
        self.fec_extended_decoder_0_0 = fec.extended_decoder(
            decoder_obj_list=dec_cc, threading=None, ann=None,
            puncpat='11', integration_period=10000)

        ##################################################
        # Connections
        ##################################################
        self.connect((self.fec_extended_decoder_0_0, 0), (self, 0))
        self.connect((self, 0), (self.fec_extended_decoder_0_0, 0))

    def get_dec_cc(self):
        return self.dec_cc

    def set_dec_cc(self, dec_cc):
        self.dec_cc = dec_cc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import LinearAdapter


Beacon = Struct(
    Const(b'\x6c'),
    Const(b'\xa3'),
    'time' / Int32ul,
    'reboots' / Int32ul,
    'rtc_val' / Int32ul,
    'bat_charge_in' / LinearAdapter(1/(2.5/(4096*20*0.05)), Int16ub),
    'bat_charge_out' / LinearAdapter(1/(2.5/(4096*20*0.033)), Int16ub),
    'bat_voltage' / LinearAdapter(1/(2.5/4096*((124+27.4)/27.4)), Int16ub),
    'supply_5V' / LinearAdapter(1/(2.5/4096*((30.1+18.2)/18.2)), Int16ub),
    'supply_3V3' / LinearAdapter(1/(2.5/4096*((18.2+18.2)/18.2)), Int16ub),
    'pcu_total_curr' / LinearAdapter(1/(2.5/(4096*20*1)), Int16ub),
    'solar_curr' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub)[6],
    'solar_total_v' / LinearAdapter(1/(2.5/4096*(30.1+18.2)/18.2), Int16ub),
    'vcc_curr' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub)[4],
    'vcc_curr2' / LinearAdapter(1/(2.5/(4096*20*0.05)), Int16ub)[4],
    'ss_total_curr' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub),
    'eeprom_curr1' / LinearAdapter(1/(2.5/(4096*20*0.2)), Int16ub),
    'eeprom_curr2' / LinearAdapter(1/(2.5/(4096*20*1)), Int16ub),
    'ext_adc_curr' / LinearAdapter(1/(2.5/(4096*20*1)), Int16ub)[4],
    'rtc_curr' / LinearAdapter(1/(2.5/(4096*20*1)), Int16ub),
    'charger_dcdc_v' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub),
    'system_v' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub),
    'obc_curr' / LinearAdapter(1/(2.5/(4096*20*0.1)), Int16ub),
    'switches' / Int8ub[3],
    'reserved1' / Int8ub,
    'battery_temp' / Int16sl,
    'schedule' / Int8ub,
    'reserved2' / Bytes(10),
    'mode' / Int8ub,
    'filler' / Bytes(10),
    'crc' / Int16ub)

dstar_one = Struct(
    'control' / Bytes(2),
    'beacon' / Beacon
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital
import pmt

from ... import pdu_head_tail
from ... import sx12xx_check_crc, sx12xx_packet_crop
from ... import reflect_bytes
from ...hier.pn9_scrambler import pn9_scrambler
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...utils.options_block import options_block


_syncword = '01010101010101010001001000010010'


class fossasat_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe FOSSASAT-1/2 custom framing

    This framing is based on Semtech SX126x transceiver
    with PN9 scrambler, bit order swapping and a CRC-16.

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'fossasat_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.slicer = digital.binary_slicer_fb()
        self.sync = sync_to_pdu_packed(
            packlen=255, sync=_syncword, threshold=syncword_threshold)
        self.reflect_1 = reflect_bytes()
        self.scrambler = pn9_scrambler()
        self.reflect_2 = reflect_bytes()
        self.crop = sx12xx_packet_crop(crc_len=2)
        self.crc = sx12xx_check_crc(
            verbose=self.options.verbose_crc, initial=0x1D0F, final=0xFFFF)
        self.remove_length = pdu_head_tail(3, 1)

        self.connect(self, self.slicer, self.sync)
        self.msg_connect((self.sync, 'out'), (self.reflect_1, 'in'))
        self.msg_connect((self.reflect_1, 'out'), (self.scrambler, 'in'))
        self.msg_connect((self.scrambler, 'out'), (self.reflect_2, 'in'))
        self.msg_connect((self.reflect_2, 'out'), (self.crop, 'in'))
        self.msg_connect((self.crop, 'out'), (self.crc, 'in'))
        self.msg_connect((self.crc, 'ok'), (self.remove_length, 'in'))
        self.msg_connect((self.remove_length, 'out'), (self, 'out'))

    _default_sync_threshold = 0

    @classmethod
    def add_options(cls, parser):
        """
        Adds FOSSASAT deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_crc', action='store_true', help='Verbose CRC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017, 2018, 2019, 2020 Daniel Estevez <daniel@destevez.net>
# Copyright 2021 The Regents of the University of Colorado
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import copy
from datetime import datetime
from construct import Adapter, BitsInteger, BitStruct, Container, Enum, \
                      Flag, GreedyBytes, If, Int8ub, Int16ub, Int32ub, \
                      Padding, RawCopy, Struct, Switch
from .ax25 import Header
from .cute_bct_fsw import cute_bct_fsw
from .cute_bct_soh import cute_bct_soh
from .cute_pld import cute_pld_sw_stat

PrimaryHeader = BitStruct(
    'ccsds_version' / BitsInteger(3),
    'packet_type' / Flag,
    'secondary_header_flag' / Flag,
    'is_stored_data' / Flag,
    'APID' / BitsInteger(10),
    'grouping_flag' / Enum(BitsInteger(2), GRP_MIDDLE=0, GRP_BEGIN=1,
                           GRP_END=2, GRP_FIRST_AND_LAST=3),
    'sequence_count' / BitsInteger(14),
    'packet_length' / BitsInteger(16)
    )


class TimeAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return Container()

    def _decode(self, obj, context, path=None):
        offset = datetime(2000, 1, 1, 12) - datetime(1970, 1, 1)
        return (datetime.utcfromtimestamp(obj.time_stamp_seconds) + offset)


SecondaryHeaderRaw = Struct(
    'time_stamp_seconds' / Int32ub,
    'sub_seconds' / Int8ub,
    Padding(1)
)

SecondaryHeader = TimeAdapter(
    SecondaryHeaderRaw
)

cute_ax25_packet_header = RawCopy(Struct(
    'ax25_header' / Header,
    'primary_header' / PrimaryHeader,
    'secondary_header' / If(
        lambda c: c.primary_header.secondary_header_flag,
        SecondaryHeader
    )
))

cute_ax25_packet_fragment = Struct(
    'header' / cute_ax25_packet_header,
    'packet' / GreedyBytes
)

cute_ax25_packet_complete = Struct(
    'ax25_header' / Header,
    'primary_header' / PrimaryHeader,
    'secondary_header' / If(
        lambda c: c.primary_header.secondary_header_flag,
        SecondaryHeader
    ),
    'packet' / Switch(
        lambda c: (c.primary_header.APID),
        {
            (0x55): cute_bct_fsw,
            (0x56): cute_bct_soh,
            (0x1FF): cute_pld_sw_stat
        }
    )
)


class FswParserState(Enum):
    SEARCH_START = 0
    SEARCH_MIDDLE = 1


class CUTE:
    """Telemetry parser for the Colorado Ultraviolet Transit Experiment (CUTE)

    This is a stateful parser that reassembles flight software packets which
    span multiple AX.25 frames.
    """
    def __init__(self):
        self.frames = []
        self.state = FswParserState.SEARCH_START

    # Parses a reassembled frame
    def process(self):
        reassembled_packet = b''.join([self.frames[0].header.data] +
                                      [frame.packet for frame in self.frames])
        self.frames = []
        return cute_ax25_packet_complete.parse(reassembled_packet)

    def parse(self, packet):
        data = cute_ax25_packet_fragment.parse(packet)
        if data is None:
            return None
        primary_header = data.header.value.primary_header

        self.frames.append(data)

        if self.state == FswParserState.SEARCH_START:

            if primary_header.grouping_flag == 'GRP_MIDDLE':
                # This means we dropped/missed a frame
                self.frames = []
                return None

            elif primary_header.grouping_flag == 'GRP_BEGIN':
                self.state = FswParserState.SEARCH_MIDDLE
                return None

            elif primary_header.grouping_flag == 'GRP_END':
                # This means we dropped/missed a frame
                self.frames = []
                return None

            elif primary_header.grouping_flag == 'GRP_FIRST_AND_LAST':
                return self.process()

        elif self.state == FswParserState.SEARCH_MIDDLE:

            if primary_header.grouping_flag == 'GRP_BEGIN' or \
               primary_header.grouping_flag == 'GRP_FIRST_AND_LAST':
                # This means we dropped/missed a frame
                self.state = FswParserState.SEARCH_START
                self.frames = []
                return None

            prev_header = self.frames[-2].header.value.primary_header
            if primary_header.sequence_count != \
               (prev_header.sequence_count + 1) % (1 << 14):
                # A frame was dropped.
                self.state = FswParserState.SEARCH_START
                self.frames = []
                return None

            if primary_header.grouping_flag == 'GRP_MIDDLE':
                return None

            elif primary_header.grouping_flag == 'GRP_END':
                self.state = FswParserState.SEARCH_START
                return self.process()


cute_70cm = CUTE()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# This contains code taken from
# https://github.com/PW-Sat2/SimpleUploader-radio.pw-sat.pl
# That code is licenced under the following terms:

# MIT License
#
# Copyright (c) 2017 SoftwareMill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
import datetime
import json

from gnuradio import gr
import numpy
import pmt

from . import hdlc


class pwsat2_submitter(gr.basic_block):
    """docstring for block pwsat2_submitter"""
    def __init__(self, credentials_file, initialTimestamp):
        gr.basic_block.__init__(
            self,
            name="pwsat2_submitter",
            in_sig=[],
            out_sig=[])
        self.requests = __import__('requests')

        self.baseUrl = 'http://radio.pw-sat.pl'
        self.headers = {'content-type': 'application/json'}

        dtformat = '%Y-%m-%d %H:%M:%S'
        self.initialTimestamp = (
            datetime.datetime.strptime(initialTimestamp, dtformat)
            if initialTimestamp != '' else None)
        self.startTimestamp = datetime.datetime.utcnow()

        self.authenticate(credentials_file)

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def authenticate(self, credentials_path):
        try:
            credentials = self.loadCredentials(credentials_path)
        except (ValueError, IOError) as e:
            print('Could not load credentials for', self.baseUrl)
            print(e)
            self.cookies = None
            return

        url = self.baseUrl+'/api/authenticate'
        response = self.requests.post(url,
                                      data=json.dumps(credentials),
                                      headers=self.headers)
        if response.status_code == 200:
            self.cookies = response.cookies
        else:
            print('Could not authenticate to PW-Sat2 server')
            print('Reply:', response.text)
            print('HTTP code', response.status_code)
            self.cookies = None

    def loadCredentials(self, path):
        with open(path) as f:
            credentials = json.load(f)
        return credentials

    def putPacket(self, frame, timestamp):
        if self.cookies is None:
            print('Not uploading packet to',
                  self.baseUrl,
                  'as we are not authenticated')
            return
        url = self.baseUrl+'/communication/frame'

        timestamp = (timestamp - datetime.datetime(1970, 1, 1)).total_seconds()
        timestamp = int(timestamp * 1000)
        payload = {'frame': str(base64.b64encode(frame), encoding='ascii'),
                   'timestamp': timestamp,
                   'traffic': 'Rx'}

        response = self.requests.put(url,
                                     data=json.dumps(payload),
                                     headers=self.headers,
                                     cookies=self.cookies)
        return response.text

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return

        data = bytearray(pmt.u8vector_elements(msg))
        crc = hdlc.crc_ccitt(data)
        data.append(crc & 0xff)
        data.append((crc >> 8) & 0xff)

        frame = bytes(data)

        now = datetime.datetime.utcnow()
        timestamp = (now - self.startTimestamp + self.initialTimestamp
                     if self.initialTimestamp else now)

        response = self.putPacket(frame, timestamp)
        if response:
            print('Packet uploaded to', self.baseUrl, response)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *


Address = BitsInteger(5)
Port = BitsInteger(6)

CSPHeader = BitStruct(
    'priority' / BitsInteger(2),
    'source' / Address,
    'destination' / Address,
    'destination_port' / Port,
    'source_port' / Port,
    'reserved' / BitsInteger(3),
    'fragmentation' / Flag,
    'hmac' / Flag,
    'xtea' / Flag,
    'rdp' / Flag,
    'crc' / Flag
    )

csp = Struct(
    'header' / CSPHeader,
    'payload' / GreedyBytes
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2016 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy
import pmt

from .telemetry import gomx_3 as tlm


class adsb_kml(gr.basic_block):
    """docstring for block adsb_kml"""
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='adsb_kml',
            in_sig=None,
            out_sig=None)

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return

        packet = bytes(pmt.u8vector_elements(msg))
        beacon = tlm.parse(packet)

        # check that message is beacon
        if (beacon.csp_header.destination != 10
                or beacon.csp_header.destination_port != 30):
            return

        if beacon.csp_header.source != 1 or beacon.beacon_type != 0:
            return

        adsb = beacon.beacon.adsb

        print("""<Placemark>
        <name>{}</name>
        <description>Altitude: {}ft Time: {}</description>
        <styleUrl>#plane</styleUrl>
        <Point><coordinates>{},{}</coordinates></Point>
</Placemark>""".format(
             hex(adsb.last_icao),
             adsb.last_alt,
             adsb.last_time,
             adsb.last_lon if adsb.last_lon <= 180 else adsb.last_lon - 360,
             adsb.last_lat))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks

from ... import kiss_to_pdu


class kiss_file_source(gr.hier_block2):
    """
    Hierarchical block for KISS file input

    The output are PDUs with frames.

    These are read from a KISS file.

    Args:
        file: input filename (string)
        options: options from argparse
    """
    def __init__(self, file, append=False, options=None):
        gr.hier_block2.__init__(
            self,
            'kiss_file_source',
            gr.io_signature(0, 0, 0),
            gr.io_signature(0, 0, 0))
        self.message_port_register_hier_out('out')

        self.filesource = blocks.file_source(gr.sizeof_char, file)
        self.kiss = kiss_to_pdu()

        self.connect(self.filesource, self.kiss)
        self.msg_connect((self.kiss, 'out'), (self, 'out'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy
import pmt

from . import hdlc_deframer
from . import hdlc


class k2sat_deframer(gr.basic_block):
    """docstring for block k2sat_deframer"""
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='k2sat_deframer',
            in_sig=[],
            out_sig=[])

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))
        self.number = 0

        # CRC-16 table construction
        self.crc_table = list()
        for i in range(256):
            tmp = 0
            if i & 1:
                tmp ^= 0x1021
            if i & 2:
                tmp ^= 0x2042
            if i & 4:
                tmp ^= 0x4084
            if i & 8:
                tmp ^= 0x8108
            if i & 16:
                tmp ^= 0x1231
            if i & 32:
                tmp ^= 0x2462
            if i & 64:
                tmp ^= 0x48C4
            if i & 128:
                tmp ^= 0x9188
            self.crc_table.append(tmp)

    def check_packet(self, packet):
        data = b'\x7e' + packet  # add 0x7e HDLC flag for CRC-16 check
        checksum = 0xFFFF
        for d in data:
            checksum = (((checksum << 8) & 0xFF00)
                        ^ self.crc_table[((checksum >> 8) ^ d) & 0x00FF])
        return checksum == 0

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        # Search all packet end markers in current packet
        start = 0
        while True:
            # Find packet end marker
            idx = packet[start:].find(b'\x7e\x55\x55\x55')
            if idx == -1:
                break
            # Check if this is a valid packet
            if self.check_packet(packet[:idx]):
                ax25_packet = packet[:-2+idx]
                ax25_packet = list(packet)  # conversion to list for pybind11
                self.message_port_pub(
                    pmt.intern('out'),
                    pmt.cons(pmt.PMT_NIL,
                             pmt.init_u8vector(len(ax25_packet), ax25_packet)))
            start = idx + 2

        return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
Construct for the Primary Header of the CCSDS Telemetry Space Data Link Protocol
( CCSDS 132.0-B-2 )
'''

from construct import *

PrimaryHeader = BitStruct('transfer_frame_version_number' / BitsInteger(2),
                          'spacecraft_id' / BitsInteger(10),
                          'virtual_channel_id' / BitsInteger(3),
                          'ocf_flag' / Flag,
                          'master_channel_frame_count' / BitsInteger(8),
                          'virtual_channel_frame_count' / BitsInteger(8),
                          'transfer_frame_secondary_header_flag' / Flag,
                          'synch_flag' / Flag,
                          'packet_order_flag' / Flag,
                          'segment_length_id' / BitsInteger(2),
                          'first_header_pointer' / BitsInteger(11))

OCFTrailer = BitStruct('control_word_type' / Flag,
                       'clcw_version_number' / BitsInteger(2),
                       'status_field' / BitsInteger(3),
                       'cop_in_effect' / BitsInteger(2),
                       'virtual_channel_identification' / BitsInteger(6),
                       'rsvd_spare1' / BitsInteger(2),
                       'no_rf_avail' / Flag,
                       'no_bit_lock' / Flag,
                       'lockout' / Flag,
                       'wait' / Flag,
                       'retransmit' / Flag,
                       'farmb_counter' / BitsInteger(2),
                       'rsvd_spare2' / Flag,
                       'report_value' / BitsInteger(8))

FullPacket = Struct('primary' / PrimaryHeader,
                    'payload' / Byte[this._.size],
                    'ocftrailer' / If(this.primary.ocf_flag == 1, OCFTrailer))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import cc11xx_packet_crop, sx12xx_check_crc
from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...utils.options_block import options_block


# The syncword is actually 0x7e preceded by a preamble of
# 5 0xaa's. We use 0xaa7e as a syncword. Each byte is set
# LSB first.
_syncword = '1010101001111110'


class endurosat_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe Endurosat frames.

    These frames are formed by a preamble of 5 0xaa bytes,
    the 0x7e flag as syncword, followed by one byte indicating
    the packet length, then the packet (length 0 to 128 bytes),
    and finally a CRC-16. The CRC-16 applies to the packet and
    packet length.

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'endurosat_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.slicer = digital.binary_slicer_fb()
        self.deframer = sync_to_pdu_packed(
            packlen=131, sync=_syncword, threshold=syncword_threshold)
        self.crop = cc11xx_packet_crop(use_crc16=True)
        self.crc = sx12xx_check_crc(
            initial=0xffff, final=0, verbose=self.options.verbose_crc)

        self.connect(self, self.slicer, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.crop, 'in'))
        self.msg_connect((self.crop, 'out'), (self.crc, 'in'))
        self.msg_connect((self.crc, 'ok'), (self, 'out'))

    _default_sync_threshold = 0

    @classmethod
    def add_options(cls, parser):
        """
        Adds Endurosat deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_crc', action='store_true', help='Verbose CRC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import sys

from ..feh import FehOpener
from .filereceiver import FileReceiver


class ImageReceiver(FileReceiver):
    """
    Class to reassemble images transmitted in chunks and display them
    in real-time using feh

    This implements the generic framework. Specific protocols should
    inherit from this class and implement some FileReceiver methods
    """
    def __init__(self, path, verbose=False, display=False, fullscreen=True):
        """
        Builds a new ImageReceiver

        Args:
            path: directory where files should be stored (str or pathlib.Path)
            verbose: verbose reporting of events (bool)
            display: display image in real-time using feh (bool)
            fullscreen: run feh in fullscreen (bool)
        """
        super().__init__(path, verbose)
        self._feh = FehOpener(fullscreen) if display else None
        self._name = 'ImageReceiver'

    def filename(self, fid):
        """
        Generates a filename based on the file id

        The default implementation uses the fid.jpg as filename

        Args:
           fid: file id (usually int)
        """
        return f'{fid}.jpg'

    def _new_file(self, fid):
        """
        Creates a new file

        Args:
            fid: file id of the new file (usually int)
        Returns:
            the new file
        """
        f = super()._new_file(fid)
        f.displaying = False
        return f

    def push_chunk(self, chunk):
        """
        Processes a new chunk

        The user should call this function whenever a new chunk
        is received.

        Args:
            chunk: the file chunk (bytes)
        """
        super().push_chunk(chunk)
        if self._current_file is None:
            return
        f = self._files[self._current_file]
        if f.write_pointer >= 10*64:
            # Enough data to display
            if self._feh is not None and not f.displaying:
                f.displaying = True
                try:
                    self._feh.open(f.path)
                except Exception as e:
                    print('Unable to open feh image viewer',
                          file=sys.stderr)
                    print(e, file=sys.stderr)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
gr-satellites ccsds

This module contains construct definitions and GNU Radio blocks
to handle CCSDS Space DataLink frames and SpacePackets
'''

from .pathID_demultiplexer import pathID_demultiplexer
from .space_packet_parser import space_packet_parser
from .space_packet_primaryheader_adder import space_packet_primaryheader_adder
from .space_packet_time_stamp_adder import space_packet_time_stamp_adder
from .telecommand_parser import telecommand_parser
from .telecommand_primaryheader_adder import telecommand_primaryheader_adder
from .telemetry_ocf_adder import telemetry_ocf_adder
from .telemetry_packet_reconstruction import telemetry_packet_reconstruction
from .telemetry_parser import telemetry_parser
from .telemetry_primaryheader_adder import telemetry_primaryheader_adder
from .virtual_channel_demultiplexer import virtual_channel_demultiplexer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
Construct for the Transfer Frame Primary Header of Fig
the CCSDS TC Space Data Link Protocol ( CCSDS 232.0-B-3 )
'''

from construct import *

PrimaryHeader = BitStruct('transfer_frame_version' / BitsInteger(2),
                          'bypass' / Flag,
                          'control' / Flag,
                          'RSVD_spare' / BitsInteger(2),
                          'spacecraft_id' / BitsInteger(10),
                          'virtual_channel_id' / BitsInteger(6),
                          'frame_length' / BitsInteger(10),
                          'frame_sequence_number' / BitsInteger(8))

FullPacket = Struct('primary' / PrimaryHeader,
                    'pyaload' / Byte[this.primary.frame_length - 5])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital, fec

from ... import decode_rs
from ...hier.ccsds_descrambler import ccsds_descrambler
from ...hier.sync_to_pdu_soft import sync_to_pdu_soft
from ...usp import usp_ax25_crop, usp_pls_crop
from ...utils.options_block import options_block


_syncword = '0101000001110010111101100100101100101101100100001011000111110101'


class usp_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the Unified SPUTNIX Protocol (USP)

    This framing is based on the CCSDS concatenated framing, but is
    optimized for variable frame size and borrows some ideas from
    DVB-S2.

    The description of the protocol can be found in
    https://sputnix.ru/tpl/docs/amateurs/
    USP%20protocol%20description%20v1.04.pdf

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'usp_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.deframer = sync_to_pdu_soft(
            packlen=4144, sync=_syncword, threshold=syncword_threshold)
        self.pls = usp_pls_crop()
        self.viterbi = fec.cc_decoder.make(
            4080, 7, 2, [79, -109], 0, -1, fec.CC_TRUNCATED, False)
        self.viterbi_decoder = fec.async_decoder(
            self.viterbi, False, False, 4080//8)
        self.scrambler = ccsds_descrambler()
        self.rs = decode_rs(True, 1)
        self.crop = usp_ax25_crop()

        self.connect(self, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.pls, 'in'))
        self.msg_connect((self.pls, 'out'), (self.viterbi_decoder, 'in'))
        self.msg_connect((self.viterbi_decoder, 'out'), (self.scrambler, 'in'))
        self.msg_connect((self.scrambler, 'out'), (self.rs, 'in'))
        self.msg_connect((self.rs, 'out'), (self.crop, 'in'))
        self.msg_connect((self.crop, 'out'), (self, 'out'))

    _default_sync_threshold = 13

    @classmethod
    def add_options(cls, parser):
        """
        Adds USP deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
gr-satellites components

These components are the basic high-level building blocks of gr-satellites.
'''

__all__ = ['deframers', 'demodulators', 'datasinks', 'transports']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks


if gr.api_version() != '9':
    from gnuradio import network

from ...utils.options_block import options_block
from ...grtypes import byte_t


class codec2_udp_sink(gr.hier_block2, options_block):
    """
    Hierarchical block for Codec2 UDP output

    The input are PDUs with Codec2 data

    The Codec2 data is sent to by PDU, a single frame (7 bytes)
    per packet for lowest latency.

    Args:
      ip: Detination IP (string)
      port: Destination UDP port (int)
    """
    def __init__(self, ip=None, port=None, options=None):
        gr.hier_block2.__init__(
            self,
            'codec2_udp_sink',
            gr.io_signature(0, 0, 0),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)
        self.message_port_register_hier_in('in')

        if ip is None:
            ip = self.options.codec2_ip
        if port is None:
            port = self.options.codec2_port

        self.pdu2tag = blocks.pdu_to_tagged_stream(byte_t, 'packet_len')
        payload_bytes = 7
        # The UDP sink has been moved in GNU Radio 3.10
        if gr.api_version() == '9':
            self.udp = udp_sink(gr.sizeof_char*1,
                                ip, port, payload_bytes, False)
        else:
            udp_header = 0  # no header
            # The new UDP sink requires at least 8 bytes of payload.
            # We use 14 bytes, which is 2 codec2 frames.
            payload_bytes = 14
            self.udp = network.udp_sink(gr.sizeof_char, 1,
                                        ip, port, udp_header,
                                        payload_bytes, False)
        self.msg_connect((self, 'in'), (self.pdu2tag, 'pdus'))
        self.connect(self.pdu2tag, self.udp)

    _default_ip = '127.0.0.1'
    _default_port = 7000

    @classmethod
    def add_options(cls, parser):
        """
        Adds Codec2 UDP output specific options to the argparse parser
        """
        parser.add_argument(
            '--codec2_ip', type=str, default=cls._default_ip,
            help='Codec2 output IP [default=%(default)r]')
        parser.add_argument(
            '--codec2_port', type=int, default=cls._default_port,
            help='Codec2 output UDP port [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import pmt

from .ao40_uncoded_crc import crc


class check_ao40_uncoded_crc(gr.basic_block):
    """docstring for block check_ao40_uncoded_crc"""
    def __init__(self, verbose):
        gr.basic_block.__init__(
            self,
            name='check_ao40_uncoded_crc',
            in_sig=[],
            out_sig=[])

        self.verbose = verbose

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = pmt.u8vector_elements(msg)
        if crc(packet) == 0:
            if self.verbose:
                print('CRC OK')
            self.message_port_pub(pmt.intern('ok'), msg_pmt)
        else:
            if self.verbose:
                print('CRC failed')
            self.message_port_pub(pmt.intern('fail'), msg_pmt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, digital

from ... import (
    nrzi_decode, hdlc_deframer, pdu_head_tail, decode_rs, pdu_length_filter)
from ...grpdu import pdu_to_tagged_stream, tagged_stream_to_pdu
from ...grtypes import byte_t
from ...utils.options_block import options_block


class ops_sat_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe OPS-SAT AX.25 + Reed-Solomon.

    The input is a float stream of soft symbols. The output are PDUs
    with Reed-Solomon decoded data.

    The input should be NRZ-I encoded and G3RUH scrambled.

    Args:
        options: Options from argparse
    """
    def __init__(self, options=None):
        gr.hier_block2.__init__(
            self,
            'ops_sat_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        self.slicer = digital.binary_slicer_fb()
        self.nrzi = nrzi_decode()
        self.descrambler = digital.descrambler_bb(0x21, 0, 16)
        self.deframer = hdlc_deframer(False, 10000)  # we skip CRC-16 check
        self.strip = pdu_head_tail(3, 16)

        # CCSDS descrambler
        self.pdu2tag = pdu_to_tagged_stream(byte_t, 'packet_len')
        self.unpack = blocks.packed_to_unpacked_bb(1, gr.GR_MSB_FIRST)
        self.scramble = digital.additive_scrambler_bb(
            0xA9, 0xFF, 7, count=0, bits_per_byte=1,
            reset_tag_key='packet_len')
        self.pack = blocks.unpacked_to_packed_bb(1, gr.GR_MSB_FIRST)
        self.tag2pdu = tagged_stream_to_pdu(byte_t, 'packet_len')

        # Prevents passing codewords of incorrect size to the RS decoder
        self.length_filter = pdu_length_filter(33, 255)
        self.fec = decode_rs(False, 1)

        self.connect(self, self.slicer, self.nrzi, self.descrambler,
                     self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.strip, 'in'))
        self.msg_connect((self.strip, 'out'), (self.pdu2tag, 'pdus'))
        self.connect(self.pdu2tag, self.unpack, self.scramble, self.pack,
                     self.tag2pdu)
        self.msg_connect((self.tag2pdu, 'pdus'), (self.length_filter, 'in'))
        self.msg_connect((self.length_filter, 'out'), (self.fec, 'in'))
        self.msg_connect((self.fec, 'out'), (self, 'out'))

    @classmethod
    def add_options(cls, parser):
        """
        Adds OPS-SAT deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--verbose_rs', action='store_true', help='Verbose RS decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks

from ... import submit, funcube_submit
from ... import pwsat2_submitter, bme_submitter, pdu_to_kiss
from ...utils.options_block import options_block


class telemetry_submit(gr.hier_block2, options_block):
    """
    Hierarchical block for telemetry submission

    The input are PDUs with frames

    These are submitted to a telemetry server

    Args:
        server: 'SatNOGS', 'FUNcube', 'PWSat', 'BME' or 'SIDS' (string)
        norad: NORAD ID (int)
        port: TCP port to connect to (used by HIT) (str)
        url: SIDS URL (used by SIDS) (str)
        config: configuration file from configparser
        options: options from argparse
    """
    def __init__(self, server, norad=None, port=None, url=None,
                 config=None, options=None):
        gr.hier_block2.__init__(
            self,
            'telemetry_submit',
            gr.io_signature(0, 0, 0),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_in('in')

        if server in ['SatNOGS', 'SIDS']:
            if server == 'SatNOGS':
                url = 'https://db.satnogs.org/api/telemetry/'
            initial_timestamp = getattr(self.options, 'start_time', '')
            self.submit = submit(url, norad,
                                 config['Groundstation']['callsign'],
                                 float(config['Groundstation']['longitude']),
                                 float(config['Groundstation']['latitude']),
                                 initial_timestamp)
        elif server == 'FUNcube':
            url = 'http://data.amsat-uk.org'
            self.submit = funcube_submit(
                url, config['FUNcube']['site_id'],
                config['FUNcube']['auth_code'])
        elif server == 'PWSat':
            self.submit = pwsat2_submitter(
                config['PW-Sat2']['credentials_file'], '')
        elif server == 'BME':
            satellites = {44830: 'atl1', 44832: 'smogp', 47964: 'smog1'}
            satellite = satellites[norad]
            self.submit = bme_submitter(
                config['BME']['user'], config['BME']['password'], satellite)
        elif server == 'HIT':
            try:
                self.tcp = blocks.socket_pdu(
                    'TCP_CLIENT', '127.0.0.1', port, 10000, False)
            except RuntimeError as e:
                print('Could not connect to telemetry proxy:', e)
                print('Disabling telemetry submission...')
                return
            self.submit = pdu_to_kiss(control_byte=False)
            self.msg_connect((self.submit, 'out'), (self.tcp, 'pdus'))
        else:
            raise ValueError('Unsupported telemetry server')

        self.msg_connect((self, 'in'), (self.submit, 'in'))

    @classmethod
    def add_options(cls, parser):
        """
        Adds telemetry submit specific options to the argparse parser
        """
        parser.add_argument(
            '--start_time', type=str, default='',
            help='Recording start timestamp')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
from datetime import datetime
from datetime import timedelta
import construct
import pmt
import array
from . import space_packet

class space_packet_time_stamp_adder(gr.basic_block):
    """
    Time Stamp Adder (CCSDS 301.0-B-4)
    --- The user should study the time code formats book and fill only the necessary fields.

    --- The user, in general, should check the code of the preferred time format and confirm that the behavior is
    the wanted one. Great care should be taken on the variability of the size. (E.g. in ASCII A time format,
    the user should define the size of the decimal fraction of the second subfield and should change the finalHeader array.)

    --- Automated time only prints a timestamp down to microseconds

    --- When defining an epoch, the user should check that it fits into the bytes that were assigned to hold the difference.
        E.g. If the epoch "0001 January 1" is defined in CDS, then it will not be able to fit today's date into a 16-bit
        DAYS part.
    """
    def __init__(self, input_manual_automatic, time_format, pfield, pfield_extension, time_code_identification_cuc,
                 epoch_year_cuc, epoch_month_cuc, epoch_day_cuc, pfield_extension_extended,rsvd_cuc,
                 time_code_identification_cds, epoch_identification_cds, epoch_year_cds, epoch_month_cds, epoch_day_cds,
                 time_code_identification_ccs, year, month, day, hour, minute, second, microsecond, picosecond, id_time):
        gr.basic_block.__init__(self,
            name="space_packet_time_stamp_adder",
            in_sig=[],
            out_sig=[])

        ##################################################
        # Parameters
        ##################################################
        self.input_manual_automatic = input_manual_automatic
        self.time_format = time_format
        self.pfield = pfield #Checks if the P-Field will be used
        self.pfield_extension = pfield_extension #Checks if the PField will be extended
        self.time_code_identification_cuc = time_code_identification_cuc
        self.epoch_year_cuc = epoch_year_cuc
        self.epoch_month_cuc = epoch_month_cuc
        self.epoch_day_cuc = epoch_day_cuc
        self.basic_time_num_octets_cuc = id_time.basic_time_num_octets_cuc
        self.fractional_time_num_octets_cuc = id_time.fractional_time_num_octets_cuc
        self.pfield_extension_extended = pfield_extension_extended #Checks if the PField Extension will be extended
        self.additional_octets_basic_time_cuc = id_time.additional_basic_time_num_octets_cuc
        self.additional_octets_fractional_time_cuc = id_time.additional_fractional_time_num_octets_cuc
        self.rsvd_cuc = rsvd_cuc
        self.time_code_identification_cds = time_code_identification_cds
        self.epoch_identification_cds = epoch_identification_cds
        self.epoch_year_cds = epoch_year_cds
        self.epoch_month_cds = epoch_month_cds
        self.epoch_day_cds = epoch_day_cds
        self.length_of_day_cds = id_time.len_of_day
        self.length_of_submillisecond_cds = id_time.len_of_submilsecs
        self.time_code_identification_ccs = time_code_identification_ccs
        self.calendar_variation_ccs = id_time.calendar_variation
        self.number_of_subsecond_ccs = id_time.num_of_subsecs
        self.add_z_terminator = id_time.add_z
        self.ascii_dec_num = id_time.ascii_dec
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond
        self.picosecond = picosecond
        self.id_time = id_time
        if self.time_code_identification_cuc == 1:
            self.epoch_cuc = datetime(1958, 1, 1)
        else:
            self.epoch_cuc = datetime(self.epoch_year_cuc, self.epoch_month_cuc, self.epoch_day_cuc)

        if self.epoch_identification_cds == 0:
            self.epoch_cds = datetime(1958, 1, 1)
        else:
            self.epoch_cds = datetime(self.epoch_year_cds, self.epoch_month_cds, self.epoch_day_cds)

        ##################################################
        # Blocks
        ##################################################
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print()
            "[ERROR] Received invalid message type. Expected u8vector"
            return
        packet = pmt.u8vector_elements(msg)
        if self.input_manual_automatic == 1:
            self.now = datetime.utcnow()
            if self.length_of_submillisecond_cds >= 2:
                self.length_of_submillisecond_cds = 1
            if self.number_of_subsecond_ccs >= 4:
                self.length_of_susecond_ccs = 3
        else:
            self.now = datetime(self.year, self.month, self.day, self.hour, self.minute, self.second, self.microsecond)

        finalHeader = []

        if self.time_format == 0: #CUC
            basic_time = 1 + self.basic_time_num_octets_cuc
            fractional_time = self.fractional_time_num_octets_cuc

            if self.pfield == 1: #If it exists

                finalHeader.extend(array.array('B', space_packet.PFieldCUC.build(
                    dict(pfield_extension=self.pfield_extension,
                         time_code_identification=self.time_code_identification_cuc,
                         number_of_basic_time_unit_octets=self.basic_time_num_octets_cuc,
                         number_of_fractional_time_unit_octets=self.fractional_time_num_octets_cuc))).tolist())

                if self.pfield_extension == 1: #If it is extended
                    basic_time += self.additional_octets_basic_time_cuc
                    fractional_time +=self.additional_octets_fractional_time_cuc
                    finalHeader.extend(array.array('B', space_packet.PFieldCUCExtension.build(
                        dict(pfieldextension=self.pfield_extension_extended,
                             number_of_additional_basic_time_unit_octets=self.additional_octets_basic_time_cuc,
                             number_of_additional_fractional_time_unit_octets=self.additional_octets_fractional_time_cuc,
                             reserved_for_mission_definition=self.rsvd_cuc))).tolist())
            temp_diff = self.now - self.epoch_cuc
            total_basic = int(temp_diff.total_seconds())
            total_frac = int((temp_diff.total_seconds() - total_basic)*(256**fractional_time))
            finalHeader.extend(array.array('B', construct.BytesInteger(basic_time).build(total_basic)).tolist())
            finalHeader.extend(array.array('B', construct.BytesInteger(fractional_time).build(total_frac)).tolist())

        elif self.time_format == 1: #CDS
            if self.pfield == 1:
                finalHeader.extend(array.array('B', space_packet.PFieldCDS.build(dict(pfield_extension = self.pfield_extension,
                                                                                 time_code_identification = self.time_code_identification_cds,
                                                                                 epoch_identification = self.epoch_identification_cds,
                                                                                 length_of_day_segment = self.length_of_day_cds,
                                                                                 length_of_submillisecond_segment = self.length_of_submillisecond_cds))).tolist())
            days_len = 2 if self.length_of_day_cds == 0 else 3
            finalHeader.extend(
                array.array('B', construct.BytesInteger(days_len).build((self.now - self.epoch_cds).days)).tolist())
            finalHeader.extend(array.array('B', construct.Int32ub.build(self.now.microsecond/1000)).tolist())
            if self.length_of_submillisecond_cds == 1:
                finalHeader.extend(array.array('B', construct.Int16ub.build(self.now.microsecond % 1000)).tolist())
            elif self.length_of_submillisecond_cds == 2:
                finalHeader.extend(array.array('B', construct.Int32ub.build(self.picosecond)).tolist())

        elif self.time_format == 2: #CCS
            if self.pfield == 1:
                finalHeader.extend(array.array('B', space_packet.PFieldCCS.build(dict(pfield_extension = self.pfield_extension,
                                                                                 time_code_identification = self.time_code_identification_ccs,
                                                                                 calendar_variation_flag = self.calendar_variation_ccs,
                                                                                 resolution=self.number_of_subsecond_ccs))).tolist())
            finalHeader.extend(array.array('B', construct.Int16ub.build(self.now.year)).tolist())
            if self.calendar_variation_ccs == 0:
                finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.month)).tolist())
                finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.day)).tolist())
            else:
                finalHeader.extend(array.array('B', construct.Int16ub.build(self.now.timetuple().tm_yday)).tolist())

            finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.hour)).tolist())
            finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.minute)).tolist())
            finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.second)).tolist())

            if self.number_of_subsecond_ccs >= 1:
                finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.microsecond / 10 ** 4)).tolist())
            if self.number_of_subsecond_ccs >= 2:
                finalHeader.extend(
                    array.array('B', construct.Int8ub.build((self.now.microsecond / 10 ** 2) % 10 ** 2)).tolist())
            if self.number_of_subsecond_ccs >= 3:
                finalHeader.extend(array.array('B', construct.Int8ub.build(self.now.microsecond % 10 ** 2)).tolist())
            if self.number_of_subsecond_ccs >= 4:
                finalHeader.extend(
                    array.array('B', construct.Int8ub.build((self.picosecond % 10 ** 6) / 10 ** 4)).tolist())
            if self.number_of_subsecond_ccs >= 5:
                finalHeader.extend(
                    array.array('B', construct.Int8ub.build((self.picosecond % 10 ** 4) / 10 ** 2)).tolist())
            if self.number_of_subsecond_ccs >= 6:
                finalHeader.extend(array.array('B', construct.Int8ub.build(self.picosecond % 10 ** 2)).tolist())

        elif self.time_format == 3 or self.time_format == 4:
            if(self.ascii_dec_num < 0 or self.ascii_dec_num > 6):
                print("Decimals of ASCII in Time Stamp Adder block should be between 0 and 6. The number was automatically set to 1.")
                self.ascii_dec_num = 1

            if self.time_format == 3: # ASCII A
                arr = self.now.isoformat()
            else: #ASCII B
                arr = self.now.strftime("%Y-%jT%H:%M:%S.%f")

            arr = arr[: -(6 - self.ascii_dec_num)]

            if (self.add_z_terminator == 1):
                arr += 'Z'

            finalHeader= array.array('B', arr).tolist()

        else:
            print("Time Format Unknown")

        finalPacket = numpy.append(finalHeader, packet)
        finalPacket = array.array('B', finalPacket[:])
        finalPacket = pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(finalPacket), finalPacket))
        self.message_port_pub(pmt.intern('out'), finalPacket)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct.core import ConstructError

from .filereceiver import FileReceiver
from ..telemetry import smogp as tlm_smogp
from ..telemetry import smog1 as tlm_smog1


class FileReceiverSMOG(FileReceiver):
    def file_id(self, chunk):
        return (
            f'spectrum_start_{chunk.startfreq}_step_{chunk.stepfreq}'
            f'_rbw_{chunk.rbw}_measid_{chunk.measid}')

    def chunk_size(self):
        return 224

    def chunk_sequence(self, chunk):
        return chunk.pckt_index

    def chunk_data(self, chunk):
        return chunk.spectrum_data

    def file_size(self, chunk):
        return chunk.pckt_count * self.chunk_size()

    def parse_chunk(self, chunk):
        try:
            frame = self._tlm.parse(chunk)
        except ConstructError:
            return None
        return frame.payload if frame.type == 5 else None


class FileReceiverSMOGP(FileReceiverSMOG):
    _tlm = tlm_smogp


class FileReceiverSMOG1(FileReceiverSMOG):
    _tlm = tlm_smog1


smogp = FileReceiverSMOGP
smog1 = FileReceiverSMOG1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
import pmt
from . import telemetry
import array
from collections import deque

class telemetry_primaryheader_adder(gr.basic_block):
    """space_packet_parser
    docstring for block telemetry_primaryheader_adder
    """
    def __init__(self, transfer_frame_version_number, spacecraft_id, virtual_channel_id, ocf_flag, master_channel_frame_count,
                 virtual_channel_frame_count, transfer_frame_secondary_header_flag, synch_flag, packet_order_flag, segment_length_id,
                 first_header_pointer, coding, reed_solomon_concat, e, q, I, turbo, ldpc_tf, ldpc_tf_size, num_of_octets):
        gr.basic_block.__init__(self,
            name="telemetry_primaryheader_adder",
            in_sig=[],
            out_sig=[])

        ##################################################
        # Parameters
        ##################################################
        num_of_octets += 1
        self.transfer_frame_version_number = 0
        self.spacecraft_id = spacecraft_id
        self.virtual_channel_id = virtual_channel_id
        self.ocf_flag = ocf_flag
        self.master_channel_frame_count = master_channel_frame_count
        self.virtual_channel_frame_count = virtual_channel_frame_count
        self.transfer_frame_secondary_header_flag = transfer_frame_secondary_header_flag
        self.synch_flag = synch_flag
        self.packet_order_flag = packet_order_flag
        self.segment_length_id = segment_length_id
        self.first_header_pointer = 0
        self.coding = coding
        self.reed_solomon_concat = reed_solomon_concat
        self.e = e
        self.q = q
        self.I = I
        self.turbo = turbo
        self.ldpc_tf = ldpc_tf
        self.ldpc_tc_size = ldpc_tf_size
        self.num_of_octets = num_of_octets
        ##################################################
        # Variables
        ##################################################
        self.list = []
        self.size = self.calculateSize()
        ##################################################
        # Blocks
        ##################################################
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = pmt.u8vector_elements(msg)

        next_pointer = 0

        while len(packet) > 0:
            limit = self.size - len(self.list)
            if len(packet) > limit:
                if len(packet) >= limit + self.size:
                    next_pointer = 0x3ff
                else:
                    next_pointer = len(packet) - limit
            elif len(packet) == limit:
                next_pointer = 0

            self.list.extend(packet[:limit])
            packet = packet[limit:]

            if len(self.list) == self.size:
                finalPacket = numpy.append(self.calculateFinalHeader(), self.list)
                self.sendPacket(finalPacket)
                self.list = []
                self.first_header_pointer = next_pointer

    def sendPacket(self, packet):
        packet = array.array('B', packet[:])
        packet = pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(packet), packet))
        self.message_port_pub(pmt.intern('out'), packet)

    def calculateSize(self):
        if self.coding == 1 or self.coding == 2 or self.coding == 7:
            size = self.num_of_octets
        elif self.coding == 3 or self.coding == 4:
            size = (255 - 2 * self.e - self.q) * self.I

            if self.reed_solomon_concat == 0:
                if ((255-self.q)*self.I)%4!=0:
                    print("Codeblock (255-q)*I is not a multiple of 4")
        elif self.coding == 5:
            size = self.turbo
        else:
            if self.ldpc_tf < 3:
                size = self.ldpc_tc_size
            else:
                size = 892
        return size

    def calculateFinalHeader(self):
        finalHeader = array.array('B', telemetry.PrimaryHeader.build(dict(transfer_frame_version_number = self.transfer_frame_version_number,
                                                                          spacecraft_id = self.spacecraft_id,
                                                                          virtual_channel_id = self.virtual_channel_id,
                                                                          ocf_flag = self.ocf_flag,
                                                                          master_channel_frame_count = self.master_channel_frame_count,
                                                                          virtual_channel_frame_count = self.virtual_channel_frame_count,
                                                                          transfer_frame_secondary_header_flag = self.transfer_frame_secondary_header_flag,
                                                                          synch_flag = self.synch_flag,
                                                                          packet_order_flag = self.packet_order_flag,
                                                                          segment_length_id = self.segment_length_id,
                                                                          first_header_pointer = self.first_header_pointer)))
        return finalHeader



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import nrzi_decode, hdlc_deframer


class ax25_deframer(gr.hier_block2):
    """
    Hierarchical block to deframe AX.25.

    The input is a float stream of soft symbols. The output are PDUs
    with AX.25 frames.

    The input should be NRZ-I encoded and optionally G3RUH scrambled.

    Args:
        g3ruh_scrambler: use G3RUH descrambling (boolean)
        options: Options from argparse
    """
    def __init__(self, g3ruh_scrambler, options=None):
        gr.hier_block2.__init__(
            self,
            'ax25_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        self.message_port_register_hier_out('out')

        self.slicer = digital.binary_slicer_fb()
        self.nrzi = nrzi_decode()
        if g3ruh_scrambler:
            self.descrambler = digital.descrambler_bb(0x21, 0, 16)
        self.deframer = hdlc_deframer(True, 10000)

        self._blocks = [self, self.slicer, self.nrzi]
        if g3ruh_scrambler:
            self._blocks.append(self.descrambler)
        self._blocks += [self.deframer]

        self.connect(*self._blocks)
        self.msg_connect((self.deframer, 'out'), (self, 'out'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018-2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
from gnuradio import gr
import numpy
import pmt


# pycrc.py --width 16 --poly 0x1021 --reflect-in true \
# --xor-in 0xffff --reflect-out true --xor-out 0xffff \
# --algorithm=table-driven --generate=c

crc_table = [
    0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
    0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
    0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
    0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
    0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
    0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
    0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
    0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
    0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
    0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
    0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
    0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
    0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
    0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
    0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
    0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
    0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
    0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
    0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
    0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
    0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
    0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
    0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
    0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
    0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
    0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
    0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
    0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
    0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
    0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
    0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
    0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78]


def crc16_x_25(data):
    crc = 0xffff

    for byte in data:
        tbl_idx = (crc ^ byte) & 0xff
        crc = (crc_table[tbl_idx] ^ (crc >> 8)) & 0xffff

    return (crc ^ 0xffff) & 0xffff


class ngham_check_crc(gr.basic_block):
    """
    Checks the CRC-16 of an NGHam frame

    Input: An NGHam frame with a CRC-16 at the end
    Output: The NGHame frame with the CRC-16 dropped,
            classified according as to whether the CRC
            is correct.
    """
    def __init__(self, verbose):
        gr.basic_block.__init__(
            self,
            name='ngham_check_crc',
            in_sig=[],
            out_sig=[])

        self.verbose = verbose

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = pmt.u8vector_elements(msg)

        if len(packet) < 3:
            return

        packet_out = packet[:-2]
        msg_out = pmt.cons(pmt.car(msg_pmt),
                           pmt.init_u8vector(len(packet_out), packet_out))
        crc = crc16_x_25(packet_out)
        if (crc >> 8) == packet[-2] and crc & 0xff == packet[-1]:
            if self.verbose:
                print('CRC OK')
            self.message_port_pub(pmt.intern('ok'), msg_out)
        else:
            if self.verbose:
                print('CRC failed')
            self.message_port_pub(pmt.intern('fail'), msg_out)
